import math
from functools import wraps
from typing import TYPE_CHECKING, Callable, TypeVar, Union

if TYPE_CHECKING:  # pragma: no cover
    from IPython.lib.pretty import RepresentationPrinter

    from measured import (
        Dimension,
        Level,
        Logarithm,
        LogarithmicUnit,
        Measurement,
        Prefix,
        Quantity,
        Unit,
    )

SUPERSCRIPTS = {
    "-": "⁻",
    ".": ".",  # There does not seem to be a superscript '.' in Unicode yet
    **{
        str(i): v
        for i, v in enumerate(["⁰", "¹", "²", "³", "⁴", "⁵", "⁶", "⁷", "⁸", "⁹"])
    },
}
DIGITS = {v: k for k, v in SUPERSCRIPTS.items()}


def superscript(exponent: Union[int, float]) -> str:
    """Given a signed integer exponent, returns the Unicode superscript string for it

    Examples:

        >>> f"x{superscript(123)}"
        'x¹²³'
        >>> f"x{superscript(0)}"
        'x⁰'
        >>> f"x{superscript(-1)}"
        'x⁻¹'
        >>> f"x{superscript(1)}"
        'x'
    """
    if exponent == 1:
        return ""

    return "".join(SUPERSCRIPTS[c] for c in str(exponent))


def from_superscript(string: str) -> int:
    """Given a Unicode superscript string, return it as an integer."""
    return int("".join(DIGITS[c] for c in string))


M = TypeVar("M")


def mathml(function: Callable[[M], str]) -> Callable[[M], str]:
    """Wraps the given MathML-producing function to produce a self-contained
    MathML tag.  This allows for composable MathML functions that can produce both
    expressions and root MathML tag."""

    @wraps(function)
    def inner(measured_object: M) -> str:
        return f"<math>{function(measured_object)}</math>"

    return inner


def dimension_repr(dimension: "Dimension") -> str:
    """Formats the given Dimension as a Python `repr`"""
    return (
        "Dimension("
        f"exponents={dimension.exponents!r}, "
        f"name={dimension.name!r}, symbol={dimension.symbol!r}"
        ")"
    )


def dimension_str(dimension: "Dimension") -> str:
    """Formats the given Dimension as a plaintext string"""
    if dimension.symbol:
        return dimension.symbol

    return (
        "⋅".join(
            f"{fundamental.symbol}{superscript(dimension.exponents[i])}"
            for i, fundamental in enumerate(dimension._fundamental)
            if dimension.exponents[i] != 0
        )
        or "?"
    )


def dimension_pretty(
    dimension: "Dimension", pretty: "RepresentationPrinter", cycle: bool
) -> None:
    """Formats the given Dimension to the provided pretty printer"""
    with pretty.group():
        pretty.text(str(dimension))
        with pretty.group(indent=2):
            pretty.break_()
            if dimension.name:
                pretty.text(dimension.name)
                pretty.break_()
            pretty.text(repr(dimension))


def dimension_mathml(dimension: "Dimension") -> str:
    """Formats the given Dimension as a MathML expression"""
    numerator, denominator = dimension.as_ratio()

    n = (
        "<mo>⋅</mo>".join(
            (
                "<msup>"
                f"<mi>{dimension.symbol}</mi>"
                "<mn>"
                f"{numerator.exponents[i] if numerator.exponents[i] != 1 else ''}"
                "</mn>"
                "</msup>"
            )
            for i, dimension in enumerate(dimension._fundamental)
            if numerator.exponents[i] != 0
        )
        or "<mi>1</mi>"
    )

    d = "<mo>⋅</mo>".join(
        (
            "<msup>"
            f"<mi>{dimension.symbol}</mi>"
            "<mn>"
            f"{denominator.exponents[i] if denominator.exponents[i] != 1 else ''}"
            "</mn>"
            "</msup>"
        )
        for i, dimension in enumerate(dimension._fundamental)
        if denominator.exponents[i] != 0
    )

    if n and not d:
        return f"<mrow>{n}</mrow>"

    return f"<mfrac><mrow>{n}</mrow><mrow>{d}</mrow></mfrac>"


def prefix_repr(prefix: "Prefix") -> str:
    """Formats the given Prefix as a Python `repr`"""
    return f"Prefix(base={prefix.base!r}, exponent={prefix.exponent!r})"


def prefix_str(prefix: "Prefix") -> str:
    """Formats the given Prefix as a plaintext string"""
    if prefix.symbol:
        return prefix.symbol
    if prefix.exponent == 0:
        return ""
    return f"{prefix.base}{superscript(prefix.exponent)}"


def prefix_pretty(
    prefix: "Prefix", pretty: "RepresentationPrinter", cycle: bool
) -> None:
    """Formats the given Prefix to the provided pretty printer"""
    with pretty.group():
        pretty.text(str(prefix))
        with pretty.group(indent=2):
            pretty.break_()
            pretty.text(f"{prefix.base}{superscript(prefix.exponent)}")
            pretty.break_()
            pretty.text(repr(prefix))


def prefix_mathml(prefix: "Prefix") -> str:
    """Formats the given Prefix as a MathML expression"""
    if prefix.symbol:
        return f"<mi>{prefix.symbol}</mi>"

    if prefix.exponent == 0:
        return ""

    return f"<msup><mn>{prefix.base}</mn><mn>{prefix.exponent}</mn></msup>"


def unit_repr(unit: "Unit") -> str:
    """Formats the given Unit as a Python `repr`"""
    if unit.name:
        return f"Unit.named({unit.name!r})"

    return (
        "Unit("
        f"prefix={unit.prefix!r}, "
        f"factors={unit.factors!r}, "
        f"dimension={unit.dimension!r}, "
        f"name={unit.name!r}, symbol={unit.symbol!r}"
        ")"
    )


def unit_str(unit: "Unit") -> str:
    """Formats the given Unit as a plaintext string"""
    if unit.symbol:
        return unit.symbol

    # In order to handle cases like `Mega * (Meter**-1)`, which naively becomes
    # "Mm⁻¹", which looks like it should parse to `(Mega*Meter)**-1`, take this
    # unit's prefix and push it down as the prefix of the first factor, which would
    # turn `Mega * (Meter**-1)` into the correct `(Micro*Meter)**-1`.
    #
    # While it seems odd to have this in `str`, it's just a side-effect of the
    # string represeentations not having parentheses.
    first, *rest = [
        (factor.prefix, factor.symbol, exponent)
        for factor, exponent in unit.factors.items()
    ]
    prefix, symbol, exponent = first
    sign = 1 if exponent >= 0 else -1
    first = ((unit.prefix * prefix) ** sign, symbol, exponent)

    return "⋅".join(
        f"{prefix}{symbol}{superscript(exponent)}"
        for prefix, symbol, exponent in [first, *rest]
    )


def unit_format(unit: "Unit", format_specifier: str) -> str:
    """Formats the given Unit as a plaintext string, using the provided format
    specifier to control the output"""
    from measured import One

    if not format_specifier:
        return unit_str(unit)

    if format_specifier == "/":
        numerator, denominator = unit.as_ratio()
        if denominator == One:
            return unit_str(unit)
        return str(numerator) + "/" + str(denominator)

    raise ValueError(f"Unrecognized format specifier {format_specifier!r}")


def unit_pretty(unit: "Unit", pretty: "RepresentationPrinter", cycle: bool) -> None:
    """Formats the given Unit to the provided pretty printer"""
    with pretty.group():
        if unit.symbol:
            pretty.text(f"{unit.symbol} ({unit:/})")
        else:
            pretty.text(f"{unit:/}")

        with pretty.group(indent=2):
            pretty.break_()
            if unit.name:
                pretty.text(unit.name)
                pretty.break_()
            pretty.pretty(unit.dimension)
            pretty.break_()
            pretty.text(repr(unit))


def unit_mathml(unit: "Unit") -> str:
    """Formats the given Unit as a MathML expression"""
    if unit.symbol:
        return f"<mi>{unit.symbol}</mi>"

    # In order to handle cases like `Mega * (Meter**-1)`, which naively becomes
    # "Mm⁻¹", which looks like it should parse to `(Mega*Meter)**-1`, take this
    # unit's prefix and push it down as the prefix of the first factor, which would
    # turn `Mega * (Meter**-1)` into the correct `(Micro*Meter)**-1`.
    #
    # While it seems odd to have this in `str`, it's just a side-effect of the
    # string represeentations not having parentheses.
    first, *rest = [
        (unit.prefix, unit.symbol, exponent) for unit, exponent in unit.factors.items()
    ]
    prefix, symbol, exponent = first
    sign = 1 if exponent >= 0 else -1
    first = ((unit.prefix * prefix) ** sign, symbol, exponent)

    numerator = [(p, s, e) for p, s, e in [first, *rest] if e >= 0]
    denominator = [(p, s, e * -1) for p, s, e in [first, *rest] if e < 0]

    n = (
        "<mo>⋅</mo>".join(
            (
                "<msup>"
                f"<mrow>{prefix_mathml(prefix)}<mi>{symbol}</mi></mrow>"
                "<mn>"
                f"{exponent if exponent != 1 else ''}"
                "</mn>"
                "</msup>"
            )
            for prefix, symbol, exponent in numerator
        )
        or "<mi>1</mi>"
    )

    d = "<mo>⋅</mo>".join(
        (
            "<msup>"
            f"<mrow>{prefix_mathml(prefix)}<mi>{symbol}</mi></mrow>"
            "<mn>"
            f"{exponent if exponent != 1 else ''}"
            "</mn>"
            "</msup>"
        )
        for prefix, symbol, exponent in denominator
    )

    if n and not d:
        return f"<mrow>{n}</mrow>"

    return f"<mfrac><mrow>{n}</mrow><mrow>{d}</mrow></mfrac>"


def quantity_repr(quantity: "Quantity") -> str:
    """Formats the given Quantity as a Python `repr`"""
    return f"Quantity(magnitude={quantity.magnitude!r}, unit={quantity.unit!r})"


def quantity_str(quantity: "Quantity") -> str:
    """Formats the given Quantity as a plaintext string"""
    return f"{quantity.magnitude} {quantity.unit}"


def quantity_format(quantity: "Quantity", format_specifier: str) -> str:
    """Formats the given Quantity as a plaintext string, using the provided format
    specifier to control the output"""
    magnitude_format, _, unit_format = format_specifier.partition(":")
    magnitude = quantity.magnitude.__format__(magnitude_format)
    unit = quantity.unit.__format__(unit_format)
    return f"{magnitude} {unit}"


def quantity_pretty(
    quantity: "Quantity", pretty: "RepresentationPrinter", cycle: bool
) -> None:
    """Formats the given Quantity to the provided pretty printer"""
    with pretty.group():
        pretty.text(f"{quantity::/}")
        with pretty.group(indent=2):
            pretty.break_()
            pretty.pretty(quantity.magnitude)
            pretty.break_()
            pretty.pretty(quantity.unit)
            pretty.break_()
            pretty.text(repr(quantity))


def quantity_mathml(quantity: "Quantity") -> str:
    """Formats the given Quantity as a MathML expression"""
    return (
        "<mrow>"
        f"<mn>{quantity.magnitude}</mn>"
        "<mo></mo>"
        f"{unit_mathml(quantity.unit)}"
        "</mrow>"
    )


def logarithm_repr(logarithm: "Logarithm") -> str:
    """Formats the given Logarithm as a Python `repr`"""
    return (
        "Logarithm("
        f"prefix={logarithm.prefix!r}, "
        f"base={logarithm.base!r}, "
        f"name={logarithm.name!r}, "
        f"symbol={logarithm.symbol!r}"
        ")"
    )


def logarithm_str(logarithm: "Logarithm") -> str:
    """Formats the given Logarithm as a plaintext string"""
    if logarithm.symbol:
        return logarithm.symbol

    function = "ln" if logarithm.base == math.e else f"log{logarithm.base}"
    return f"{logarithm.prefix.quantify()} {function}(x/x₀)"


def logarithm_pretty(
    logarithm: "Logarithm", pretty: "RepresentationPrinter", cycle: bool
) -> None:
    """Formats the given Logarithm to the provided pretty printer"""
    with pretty.group():
        pretty.text(str(logarithm))
        with pretty.group(indent=2):
            pretty.break_()
            pretty.text(repr(logarithm))


def logarithm_mathml(logarithm: "Logarithm") -> str:
    """Formats the given Logarithm as a MathML expression"""
    if logarithm.symbol:
        return f"<mi>{logarithm.symbol}</mi>"

    prefix = (
        f"<mn>{logarithm.prefix.quantify()}</mn>" if logarithm.prefix.exponent else ""
    )
    function = "ln" if logarithm.base == math.e else f"log{logarithm.base}"
    return (
        "<mrow>"
        f"{prefix}"
        f"<mi>{function}</mi>"
        "<mo>(</mo>"
        "<mfrac>"
        f"<mi>x</mi>"
        f"<mi>x₀</mi>"
        "</mfrac>"
        "<mo>)</mo>"
        "</mrow>"
    )


def logarithmic_unit_repr(logarithmic_unit: "LogarithmicUnit") -> str:
    """Formats the given LogarithmicUnit as a Python `repr`"""
    return (
        "LogarithmicUnit("
        f"logarithm={logarithmic_unit.logarithm!r}, "
        f"reference={logarithmic_unit.reference!r}"
        ")"
    )


def logarithmic_unit_str(logarithmic_unit: "LogarithmicUnit") -> str:
    """Formats the given LogarithmicUnit as a plaintext string"""
    if logarithmic_unit.symbol:
        return logarithmic_unit.symbol

    return f"{logarithmic_unit.logarithm} of {logarithmic_unit.reference}"


def logarithmic_unit_pretty(
    logarithmic_unit: "LogarithmicUnit", pretty: "RepresentationPrinter", cycle: bool
) -> None:
    """Formats the given LogarithmicUnit to the provided pretty printer"""
    with pretty.group():
        pretty.text(str(logarithmic_unit))
        with pretty.group(indent=2):
            pretty.break_()
            pretty.text(repr(logarithmic_unit))


def logarithmic_unit_mathml(logarithmic_unit: "LogarithmicUnit") -> str:
    """Formats the given LogarithmicUnit as a MathML expression"""
    if logarithmic_unit.symbol:
        return f"<mi>{logarithmic_unit.symbol}</mi>"

    return (
        "<mrow>"
        f"{logarithm_mathml(logarithmic_unit.logarithm)}"
        "<mo>of</mo>"
        f"{quantity_mathml(logarithmic_unit.reference)}"
        "</mrow>"
    )


def level_repr(level: "Level") -> str:
    """Formats the given Level as a Python `repr`"""
    return f"Level(magnitude={level.magnitude!r}, unit={level.unit!r})"


def level_str(level: "Level") -> str:
    """Formats the given Level as a plaintext string"""
    return f"{level.magnitude} {level.unit}"


def level_pretty(level: "Level", pretty: "RepresentationPrinter", cycle: bool) -> None:
    """Formats the given Level to the provided pretty printer"""
    with pretty.group():
        pretty.text(str(level))
        with pretty.group(indent=2):
            pretty.break_()
            pretty.pretty(level.magnitude)
            pretty.break_()
            pretty.pretty(level.unit)
            pretty.break_()
            pretty.text(repr(level))


def level_mathml(level: "Level") -> str:
    """Formats the given Level as a MathML expression"""
    return (
        "<mrow>"
        f"<mn>{level.magnitude}</mn>"
        "<mo>⋅</mo>"
        f"{logarithmic_unit_mathml(level.unit)}"
        "</mrow>"
    )


def measurement_repr(measurement: "Measurement") -> str:
    """Formats the given Measurement as a Python `repr`"""
    return (
        f"Measurement("
        f"measurand={measurement.measurand!r}, "
        f"uncertainty={measurement.uncertainty.magnitude!r}"
        ")"
    )


def measurement_str(measurement: "Measurement") -> str:
    """Formats the given Measurement as a plaintext string"""
    return measurement_format(measurement, "")


def measurement_format(measurement: "Measurement", format_specifier: str) -> str:
    """Formats the given Measurement as a plaintext string, using the provided format
    specifier to control the output"""
    uncertainty_format, _, quantity_format = format_specifier.partition(":")

    style, magnitude_format = "", ""
    if uncertainty_format:
        style, magnitude_format = uncertainty_format[0], uncertainty_format[1:]

    if style in ("", "+", "±"):
        magnitude = measurement.uncertainty.magnitude.__format__(magnitude_format)
        uncertainty = f"±{magnitude}"
    elif style == "%":
        magnitude_format = magnitude_format or ".2f"
        percent = measurement.uncertainty_percent.__format__(magnitude_format)
        uncertainty = f"±{percent}%"
    else:
        raise ValueError(f"Unrecognized uncertainty style {style!r}")

    magnitude_format, _, unit_format = quantity_format.partition(":")
    magnitude = measurement.measurand.magnitude.__format__(magnitude_format)
    unit = measurement.measurand.unit.__format__(unit_format)
    return f"{magnitude}{uncertainty} {unit}"


def measurement_pretty(
    measurement: "Measurement", pretty: "RepresentationPrinter", cycle: bool
) -> None:
    """Formats the given Measurement to the provided pretty printer"""
    with pretty.group():
        pretty.text(f"{measurement:::/}")
        with pretty.group(indent=2):
            pretty.break_()
            pretty.pretty(measurement.measurand.magnitude)
            pretty.text(" ±")
            pretty.pretty(measurement.uncertainty.magnitude)
            pretty.break_()
            pretty.pretty(measurement.measurand.unit)
            pretty.break_()
            pretty.text(repr(measurement))


def measurement_mathml(measurement: "Measurement") -> str:
    """Formats the given Measurement as a MathML expression"""
    return (
        "<mrow>"
        f"<mn>{measurement.measurand.magnitude}</mn>"
        "<mo>±</mo>"
        f"<mn>{measurement.uncertainty.magnitude}</mn>"
        "<mo></mo>"
        f"{unit_mathml(measurement.measurand.unit)}"
        "</mrow>"
    )



# https://web.archive.org/web/20120204081321/http://edocket.access.gpo.gov/cfr_2004/aprqtr/21cfr101.9.htm
#
# (viii) For nutrition labeling purposes, a teaspoon means 5
# milliliters (mL), a tablespoon means 15 mL, a cup means 240 mL, 1 fl oz
# means 30 mL, and 1 oz in weight means 28 g.
# [pp 23]

from measured import approximately
from measured.si import Gram, Liter, Milli
from measured.us import Cup, FluidOunce, Ounce, Tablespoon, Teaspoon


def test_teaspoon_is_about_5mL() -> None:
    assert 1 * Teaspoon == approximately(5 * Milli * Liter, within=0.015)


def test_tablespoon_is_about_15mL() -> None:
    assert 1 * Tablespoon == approximately(15 * Milli * Liter, within=0.015)


def test_cup_is_about_240mL() -> None:
    assert 1 * Cup == approximately(240 * Milli * Liter, within=0.015)


def test_fluid_ounce_is_about_30mL() -> None:
    assert 1 * FluidOunce == approximately(30 * Milli * Liter, within=0.015)


def test_dry_ounce_is_about_28g() -> None:
    assert 1 * Ounce == approximately(28 * Gram, within=0.013)



from typing import Any

import pytest
from hypothesis import assume, given
from hypothesis.strategies import floats, integers, one_of

from measured import Length, Numeric, One, Quantity, Unit, approximately
from measured.hypothesis import quantities
from measured.si import Meter, Second


@given(value=one_of(floats(allow_nan=False, allow_infinity=False), integers()))
def test_constructing_quantities_by_multiplying_units(value: Numeric) -> None:
    assert Meter * value == value * Meter
    assert value * Meter * Meter == value * Meter**2
    assert value * Meter / Second == value * (Meter / Second)


@pytest.mark.parametrize("value", ["nope", b"nah"])
def test_construction_by_multiplication_only_with_number_types(value: Any) -> None:
    with pytest.raises(TypeError):
        Meter * value

    with pytest.raises(TypeError):
        value * Meter


@pytest.mark.parametrize("value", ["nope", b"nah"])
def test_construction_by_division_only_with_number_types(value: Any) -> None:
    with pytest.raises(TypeError):
        Meter / value

    with pytest.raises(TypeError):
        value / Meter


def test_equality_only_with_quantities() -> None:
    assert 5 * One != 5
    assert 5 * Meter != 5
    assert 5 * Meter != Meter


def test_equality_only_in_same_dimension() -> None:
    assert 5 * Meter != 5 * Second


def test_addition_only_with_quantities() -> None:
    with pytest.raises(TypeError):
        (5 * Meter) + 10  # type: ignore

    with pytest.raises(TypeError):
        10 + (5 * Meter)  # type: ignore


def test_subtraction_only_with_quantities() -> None:
    with pytest.raises(TypeError):
        (5 * Meter) - 10  # type: ignore

    with pytest.raises(TypeError):
        10 - (5 * Meter)  # type: ignore


def test_multiplication_by_scalar() -> None:
    assert 10 * (5 * Meter) == 50 * Meter


def test_division_by_scalar() -> None:
    assert (500 * Meter) / 10 == 50 * Meter


def test_exponentation_by_scalar() -> None:
    assert (5 * Meter) ** 2 == 25 * Meter**2


def test_roots() -> None:
    assert (10 * Meter).root(0) == 1 * One
    assert ((10 * Meter) ** 2).root(2) == 10 * Meter
    assert (100 * Meter**2).root(2) == 10 * Meter
    assert (64 * Meter**3).root(3) == approximately(4 * Meter)


def test_cannot_multiply_by_random_types() -> None:
    with pytest.raises(TypeError):
        "hi" * (5 * Meter)  # type: ignore

    with pytest.raises(TypeError):
        (5 * Meter) * "hi"  # type: ignore


def test_cannot_divide_by_random_types() -> None:
    with pytest.raises(TypeError):
        (5 * Meter) / "hi"  # type: ignore


def test_negative() -> None:
    assert -(5 * Meter) == -5 * Meter


def test_positive() -> None:
    assert +(5 * Meter) == +5 * Meter


def test_absolute_value() -> None:
    assert abs(-5 * Meter) == 5 * Meter


def test_ordering() -> None:
    assert (-5 * Meter) < (1 * Meter)
    assert (1 * Meter) > (-5 * Meter)
    assert (1 * Meter) >= (1 * Meter)
    assert (1 * Meter) <= (1 * Meter)


def test_ordering_only_with_quantities() -> None:
    with pytest.raises(TypeError):
        1 * Meter <= 1


def test_ordering_only_within_dimension() -> None:
    with pytest.raises(TypeError):
        1 * Meter < 1 * Second

    with pytest.raises(TypeError):
        1 * Meter <= 1 * Second

    with pytest.raises(TypeError):
        1 * Meter > 1 * Second

    with pytest.raises(TypeError):
        1 * Meter >= 1 * Second


def test_ordering_requires_types_with_a_conversion() -> None:
    Bogie = Unit.define(Length, name="bogie", symbol="bogie")

    one = 1 * Meter
    other = 1 * Bogie

    with pytest.raises(TypeError):
        one < other

    with pytest.raises(TypeError):
        one <= other

    with pytest.raises(TypeError):
        one > other

    with pytest.raises(TypeError):
        one >= other


def test_ordering_requires_types_in_the_same_dimension() -> None:
    one = 1 * Meter
    other = 1 * Second

    with pytest.raises(TypeError):
        one < other

    with pytest.raises(TypeError):
        one <= other

    with pytest.raises(TypeError):
        one > other

    with pytest.raises(TypeError):
        one >= other


def test_sorting() -> None:
    unsorted = [1 * Meter, 0 * Meter, 5 * Meter]
    assert sorted(unsorted) == [0 * Meter, 1 * Meter, 5 * Meter]


def test_approximating() -> None:
    assert 1 * Meter == approximately(1.0 * Meter)
    assert 1 * Meter == approximately(1 * Meter)
    assert 1.0 * Meter == approximately(1.0 * Meter)
    assert 1 * Meter == approximately(1.0000000001 * Meter)
    assert 1 * Meter != approximately(1.01 * Meter)
    assert 1 * Meter != approximately(1.0 * Second)


@given(quantity=quantities())
def test_repr(quantity: Quantity) -> None:
    assert (
        repr(quantity)
        == f"Quantity(magnitude={quantity.magnitude!r}, unit={quantity.unit!r})"
    )


@given(quantity=quantities())
def test_repr_roundtrips(quantity: Quantity) -> None:
    from measured import Dimension, Prefix, Unit  # noqa: F401

    assert eval(repr(quantity)) == quantity


@given(quantity=quantities())
def test_hashing(quantity: Quantity) -> None:
    assert isinstance(hash(quantity), int)

    equal = Quantity(quantity.magnitude, quantity.unit)
    assert quantity is not equal
    assert hash(quantity) == hash(equal)

    unequal = Quantity(quantity.magnitude + 1, quantity.unit)
    assert quantity is not unequal
    assume(hash(quantity.magnitude) != hash(unequal.magnitude))
    assert hash(quantity) != hash(unequal)

    assert {quantity, equal, unequal} == {quantity, unequal}
    assert {quantity: "hi"}[quantity] == "hi"
    assert {quantity: "hi"}[equal] == "hi"
    assert {quantity: "hi"}.get(unequal) is None



import pytest

from measured import Quantity, approximately
from measured.energy import (
    BritishThermalUnit,
    Calorie,
    Donkeypower,
    Horsepower,
    TonneOfTNT,
    TonOfRefrigeration,
)
from measured.si import ElectronVolt, Hour, Joule, Kilo, Watt
from measured.us import Foot, PoundForce


@pytest.mark.parametrize(
    "equivalent, within",
    [
        (4.184 * Joule, 0),
        (1.162e-6 * (Kilo * Watt) * Hour, 2e-4),
        (2.611e19 * ElectronVolt, 2e-4),
    ],
)
def test_calorie(equivalent: Quantity, within: float) -> None:
    assert 1 * Calorie == approximately(equivalent, within=within)
    assert equivalent == approximately(1 * Calorie, within=within)


@pytest.mark.parametrize(
    "equivalent, within",
    [
        (1.054350 * Kilo * Joule, 1e-6),
        (0.2931 * Watt * Hour, 8e-4),
        (252.2 * Calorie, 9e-4),
        (0.2522 * Kilo * Calorie, 9e-4),
        (778.2 * Foot * PoundForce, 8e-4),
    ],
)
def test_btus(equivalent: Quantity, within: float) -> None:
    assert 1 * BritishThermalUnit == approximately(equivalent, within=within)
    assert equivalent == approximately(1 * BritishThermalUnit, within=within)


def test_refrigeration() -> None:
    assert 1 * TonOfRefrigeration == approximately(12000 * BritishThermalUnit / Hour, 0)
    assert 1 * TonOfRefrigeration == approximately(3.51685 * Kilo * Watt, 0)
    assert 1 * TonOfRefrigeration == approximately(
        3025.97 * (Kilo * Calorie) / Hour, 2e-7
    )


@pytest.mark.parametrize(
    "equivalent, within",
    [
        (4.184e9 * Joule, 0),
        (1.0e9 * Calorie, 0),
        (3.96831e6 * BritishThermalUnit, 3e-6),
        (3.086e9 * Foot * PoundForce, 2e-5),
        (1.162e3 * Kilo * Watt * Hour, 2e-4),
    ],
)
def test_tnt_equivalent(equivalent: Quantity, within: float) -> None:
    assert 1 * TonneOfTNT == approximately(equivalent, within=within)
    assert equivalent == approximately(1 * TonneOfTNT, within=within)


def test_horsepower() -> None:
    assert 1 * Horsepower == approximately(745.69987 * Watt)


def test_donkeypower() -> None:
    assert 1 * Donkeypower == approximately(1 / 3 * Horsepower)



import pytest

from measured import Length, Quantity, approximately
from measured.si import Kilo, Meter, Micro, Milli
from measured.us import (
    Cable,
    Chain,
    Fathom,
    Foot,
    Furlong,
    Inch,
    League,
    Link,
    Mile,
    NauticalMile,
    Pica,
    Point,
    Rod,
    StatuteMile,
    SurveyFoot,
    Yard,
)


def test_point() -> None:
    assert Point.name == "point"
    assert Point.symbol == "p."
    assert Point.dimension is Length

    assert 1 * Point == (1 / 12) * Pica
    assert 1 * Point == approximately((1 / 72) * Inch)
    assert 1 * Point == (127 / 360 * Milli * Meter)
    assert 1 * Point == approximately(352.778 * Micro * Meter, within=7e-7)


def test_pica() -> None:
    assert Pica.name == "pica"
    assert Pica.symbol == "P."
    assert Pica.dimension is Length

    assert 1 * Pica == 12 * Point
    assert 1 * Pica == (1 / 6) * Inch
    assert 1 * Pica == approximately(4.233 * Milli * Meter, within=8e-5)


def test_inch() -> None:
    assert Inch.name == "inch"
    assert Inch.symbol == "in."
    assert Inch.dimension is Length

    assert 1 * Inch == 72 * Point
    assert 1 * Inch == 6 * Pica
    assert 1 * Inch == (1 / 12) * Foot
    assert 1 * Inch == (1 / 36) * Yard
    assert 1 * Inch == 25.4 * Milli * Meter


def test_foot() -> None:
    assert Foot.name == "foot"
    assert Foot.symbol == "ft."
    assert Foot.dimension is Length

    assert 1 * Foot == approximately(864 * Point)
    assert 1 * Foot == 72 * Pica
    assert 1 * Foot == 12 * Inch
    assert 1 * Foot == (1 / 3) * Yard
    assert 1 * Foot == 0.3048 * Meter


def test_yard() -> None:
    assert Yard.name == "yard"
    assert Yard.symbol == "yd."
    assert Yard.dimension is Length

    assert 1 * Yard == 36 * Inch
    assert 1 * Yard == 3 * Foot
    assert 1 * Yard == 0.9144 * Meter


def test_mile() -> None:
    assert Mile.name == "mile"
    assert Mile.symbol == "mi."
    assert Mile.dimension is Length

    assert 1 * Mile == 5280 * Foot
    assert 1 * Mile == 1760 * Yard
    assert 1 * Mile == 1.609344 * Kilo * Meter


def test_ordering() -> None:
    assert 1 * Point < 1 * Pica
    assert 1 * Pica < 1 * Inch
    assert 1 * Inch < 1 * Foot
    assert 1 * Foot < 1 * Yard
    assert 1 * Yard < 1 * Meter
    assert 1 * Kilo * Meter < 1 * Mile


def test_survey_measures() -> None:
    assert 1 * Link == 792 / 3937 * Meter
    assert 1 * Link == approximately(0.2011684 * Meter)

    assert 1 * SurveyFoot == 1200 / 3937 * Meter
    assert 1 * SurveyFoot == approximately(0.3048006 * Meter)

    assert 1 * Rod == (19800 / 3937) * Meter
    assert 1 * Rod == approximately(5.0292100 * Meter)

    assert 1 * Chain == (79200 / 3937) * Meter
    assert 1 * Chain == approximately(20.1168402 * Meter)

    assert 1 * Furlong == (792 / 3937) * Kilo * Meter
    assert 1 * Furlong == approximately(201.1684023 * Meter)

    assert 1 * StatuteMile == (6336 / 3937) * Kilo * Meter
    assert 1 * StatuteMile == approximately(1609.3472186 * Meter)

    assert 1 * League == (19008 / 3937) * Kilo * Meter
    assert 1 * League == approximately(4828.0416560 * Meter)


@pytest.mark.parametrize(
    "left, right",
    [
        (1 * Fathom, 2 * Yard),
        (1 * Fathom, 1.8288 * Meter),
        (1 * NauticalMile, 1.852 * Kilo * Meter),
    ],
)
def test_nautical_measures_equal(left: Quantity, right: Quantity) -> None:
    assert left == right, f"{left} != {right.in_unit(left.unit)}"
    assert right == left, f"{right} != {left.in_unit(right.unit)}"


@pytest.mark.parametrize(
    "left, right, within",
    [
        (1 * Cable, 0.219456 * Kilo * Meter, 1e-6),
        (1 * NauticalMile, 1.151 * Mile, 1e-3),
    ],
)
def test_nautical_measures_approximate(
    left: Quantity, right: Quantity, within: float
) -> None:
    assert left == approximately(right, within)
    assert right == approximately(left, within)



"""
Units and non-standardized conventions relating to electronics

Attributes: LogarithmicUnits

    dBW (LogarithmicUnit): Decibels of `Power`, compared to 1 `Watt`
"""
from measured import Decibel
from measured.si import Watt

# https://en.wikipedia.org/wiki/Decibel#Suffixes_and_reference_values

dBW = Decibel[1 * Watt].alias("dbW", "dbW")



import functools
import operator
from collections import defaultdict
from functools import reduce
from typing import Dict, Iterable, List, Set, Tuple, TypeVar

from measured import ic  # noqa: F401
from measured import (
    Dimension,
    FractionalDimensionError,
    Number,
    Numeric,
    One,
    Quantity,
    Unit,
)

from .compat import gcd

Ratio = Numeric
Offset = Numeric

_ratios: Dict[Unit, Dict[Unit, Ratio]] = defaultdict(dict)
_offsets: Dict[Unit, Dict[Unit, Offset]] = defaultdict(dict)


def equate(a: Quantity, b: Quantity) -> None:
    """Defines a conversion between one Unit and another, expressed as a ratio
    between the two."""

    if a.unit == b.unit and a.unit is not One:
        raise ValueError("No need to define conversions for a unit and itself")

    a = a.unprefixed()
    b = b.unprefixed()

    _ratios[a.unit][b.unit] = b.magnitude / a.magnitude
    _ratios[b.unit][a.unit] = a.magnitude / b.magnitude


def translate(scale: Unit, zero: Quantity) -> None:
    """Defines a unit as a scale starting from the given zero point in another
    unit"""
    if scale == zero.unit:
        raise ValueError("No need to define conversions for a unit and itself")

    degree = zero.unit
    offset = zero.magnitude

    _ratios[degree][scale] = 1
    _ratios[scale][degree] = 1

    _offsets[degree][scale] = -offset
    _offsets[scale][degree] = +offset


class ConversionNotFound(ValueError):
    pass


def convert(quantity: Quantity, other_unit: Unit) -> Quantity:
    """Converts the given quantity into another unit, if possible"""
    if quantity.unit.dimension != other_unit.dimension:
        raise ConversionNotFound(
            "No conversion from "
            f"{quantity.unit} ({quantity.unit.dimension}) to "
            f"{other_unit} ({other_unit.dimension})"
        )

    this = quantity.unprefixed()

    plan = _plan_conversion(quantity.unit, other_unit)

    magnitude = this.magnitude

    for ratio, path, exponent in plan:
        magnitude *= ratio
        for scale, offset, _ in path:
            magnitude *= scale**exponent
            magnitude += offset

    return Quantity(magnitude, other_unit)


Exponent = int
Path = List[Tuple[Ratio, Offset, Unit]]
RoughPlan = List[Tuple[Ratio, Unit, Unit, Exponent]]
Plan = List[Tuple[Ratio, Path, Exponent]]


@functools.lru_cache(maxsize=None)
def _plan_conversion(start: Unit, end: Unit) -> Plan:
    unprefixed = end.quantify()
    plan: RoughPlan = [(1 / unprefixed.magnitude, One, One, 1)]

    start_factors = _splat(start)
    end_factors = _splat(end)

    direct_path = _find_path(start, end)
    if direct_path:
        return _inline_paths(plan) + [(1, direct_path, 1)]

    plan += [
        (ratio, end, start, exponent)
        for ratio, start, end, exponent in _replace_factors(start_factors)
    ]
    plan += [
        (1 / ratio, start, end, exponent)
        for ratio, start, end, exponent in _replace_factors(end_factors)
    ]

    plan += _match_factors(start_factors, end_factors)
    plan += [
        (1, end, start, exponent)
        for ratio, start, end, exponent in _match_factors(end_factors, start_factors)
    ]

    plan += _cancel_factors(end_factors)
    plan += _cancel_factors(start_factors, invert=True)

    assert not start_factors
    assert not end_factors

    return _inline_paths(plan)


def _inline_paths(plan: List[Tuple[Ratio, Unit, Unit, Exponent]]) -> Plan:
    inlined: Plan = []
    for ratio, start, end, exponent in plan:
        path = _find_path(start, end)
        if not path:
            raise ConversionNotFound(f"No conversion from {start} to {end}")
        inlined.append((ratio, path, exponent))
    return inlined


def _replace_factors(factors: Dict[Dimension, List[Unit]]) -> RoughPlan:
    plan: RoughPlan = []

    # Continue looking for replacements until the plan stops changing
    previous_plan_size = -1
    while len(plan) != previous_plan_size:
        previous_plan_size = len(plan)

        replacements: List[Tuple[Dimension, Unit, Unit]] = []

        for dimension, units in factors.items():
            # Only try this for units in higher/derived dimensions
            if sum(abs(e) for e in dimension.exponents) <= 1:
                continue

            for unit in units:
                alternatives = sorted(
                    _ratios[unit].keys(),
                    key=lambda u: len(u.factors) + sum(u.factors.values()),
                    reverse=True,
                )
                for alternative in alternatives:
                    # Consider this a better alternative if it has more factors, or it
                    # has factors with higher exponents, and will thus "splat" out into
                    # a larger number of smaller/more fundamental units
                    has_more_factors = len(alternative.factors) > len(unit.factors)
                    has_smaller_factors = sum(alternative.factors.values()) > sum(
                        unit.factors.values()
                    )

                    if has_more_factors or has_smaller_factors:
                        replacements.append((dimension, unit, alternative))
                        break

        for dimension, unit, alternative in replacements:
            overall_sign = 1
            if not unit.dimension.is_factor(dimension):
                assert (unit**-1).dimension.is_factor(dimension)
                overall_sign = -1

            ratio = _ratios[unit][alternative]

            _clean_remove(factors, dimension, unit)
            for unit, exponent in alternative.factors.items():
                unit_sign = -1 if exponent < 0 else 1
                unit_dimension = unit.dimension ** (unit_sign * overall_sign)
                if unit_dimension not in factors:
                    factors[unit_dimension] = []
                factors[unit_dimension].extend([unit] * abs(exponent))

            plan.append((ratio**overall_sign, One, One, 1))

    return plan


def _match_factors(
    start_factors: Dict[Dimension, List[Unit]],
    end_factors: Dict[Dimension, List[Unit]],
) -> RoughPlan:

    plan: RoughPlan = []

    dimensions_to_match = _by_complex_first(
        dimension for dimension, factors in end_factors.items() for _ in factors
    )

    for end_dimension in dimensions_to_match:
        # go see if you can make a complete factor of end_dimension out of the
        # dimensions available in start_factors
        dimension_factors: List[Dimension] = []
        dimensions_to_check = _by_complex_first(
            dimension for dimension, factors in start_factors.items() for _ in factors
        )

        remaining = end_dimension
        while dimensions_to_check:
            start_dimension = dimensions_to_check.pop(0)
            if start_dimension.is_factor(remaining):
                dimension_factors.append(start_dimension)
                remaining /= start_dimension

            if remaining is Number:
                break

        if not dimension_factors:
            continue

        discovered_dimension = reduce(operator.mul, dimension_factors)
        if discovered_dimension is not end_dimension:
            continue

        combined_start_factor: Unit = reduce(
            operator.mul,
            [_clean_pop(start_factors, d) for d in dimension_factors],
        )
        end_factor = _clean_pop(end_factors, end_dimension)

        # TODO: this doesn't seem right in light of complex units with mixed exponents
        exponent = -1 if any(e < 0 for e in end_dimension.exponents) else 1

        plan.append((1, combined_start_factor, end_factor, exponent))

    return plan


def _cancel_factors(
    factors: Dict[Dimension, List[Unit]], invert: bool = False
) -> RoughPlan:
    plan: RoughPlan = []

    for dimension in list(factors):
        exponent = -1 if any(e < 0 for e in dimension.exponents) else 1
        inverse = dimension**-1
        while dimension in factors and inverse in factors:
            end_factor = _clean_pop(factors, dimension)
            if dimension is inverse:
                continue
            start_factor = _clean_pop(factors, inverse)

            if invert:
                plan.append((1, start_factor, end_factor, -exponent))
                plan.append((1, end_factor, end_factor, exponent))
            else:
                plan.append((1, start_factor, end_factor, exponent))
                plan.append((1, start_factor, start_factor, -exponent))

    return plan


def _splat(unit: Unit) -> Dict[Dimension, List[Unit]]:
    splatted: Dict[Dimension, List[Unit]] = defaultdict(list)

    for factor, exponent in unit.factors.items():
        if exponent < 0:
            splatted[factor.dimension**-1].extend([factor] * abs(exponent))
        else:
            splatted[factor.dimension].extend([factor] * exponent)

    return dict(splatted)


def _by_complex_first(dimensions: Iterable[Dimension]) -> List[Dimension]:
    return sorted(
        dimensions, key=lambda d: sum(abs(e) for e in d.exponents), reverse=True
    )


K = TypeVar("K")
V = TypeVar("V")


def _clean_pop(dictionary: Dict[K, List[V]], key: K) -> V:
    value = dictionary[key].pop(0)
    if not dictionary[key]:
        dictionary.pop(key)
    return value


def _clean_remove(dictionary: Dict[K, List[V]], key: K, item: V) -> None:
    dictionary[key].remove(item)
    if not dictionary[key]:
        dictionary.pop(key)


@functools.lru_cache(maxsize=None)
def _find_path(
    start: Unit,
    end: Unit,
) -> Path:
    return _find_path_recursive(start, end, visited=set())


def _find_path_recursive(
    start: Unit,
    end: Unit,
    visited: Set[Unit],
) -> Path:

    if start is end:
        return [(1, 0, end)]

    if start in visited:
        return []
    else:
        visited.add(start)

    exponent, start, end = _reduce_dimension(start, end)

    best_path: Path = []

    for intermediate, scale in _ratios[start].items():
        offset = _offsets[start].get(intermediate, 0)
        if intermediate == end:
            return [(scale**exponent, offset**exponent, end**exponent)]

        path = _find_path_recursive(intermediate, end, visited=visited)
        if not path:
            continue

        path = [(scale, offset, intermediate)] + list(path)
        path = [
            (scale**exponent, offset**exponent, unit**exponent)
            for scale, offset, unit in path
        ]
        if not best_path or len(path) < len(best_path):
            best_path = path

    if best_path:
        return best_path

    return []


def _reduce_dimension(start: Unit, end: Unit) -> Tuple[int, Unit, Unit]:
    """Reduce the dimension of the given units to their lowest common exponents"""
    assert start.dimension is end.dimension, (
        f"{start} ({start.dimension}) and {end} ({end.dimension}) measure "
        "different dimensions"
    )

    if start.dimension is Number:
        return 1, start, end

    exponent = gcd(*start.dimension.exponents)

    try:
        start_root = start.root(exponent)
        end_root = end.root(exponent)
    except FractionalDimensionError:
        return 1, start, end

    return exponent, start_root, end_root



from decimal import Decimal
from typing import Any

import pytest
from pytest import approx

from measured import Decibel, Measurement, Quantity, approximately
from measured.si import Meter, Second
from measured.us import Foot


def test_addition() -> None:
    # adapted from https://www.statisticshowto.com/statistics-basics/error-propagation/
    waistband = Measurement(0.88 * Meter, 0.03)
    pant_length = Measurement(1.12 * Meter, 0.04)
    height = pant_length + waistband
    assert height.measurand == 2.0 * Meter
    assert height.uncertainty == 0.05 * Meter
    assert height.uncertainty_ratio == 0.025


def test_addition_with_quantity() -> None:
    waistband = Measurement(0.88 * Meter, 0.03)
    pant_length = 1.12 * Meter
    height = pant_length + waistband
    assert height.measurand == 2.0 * Meter
    assert height.uncertainty == 0.03 * Meter
    assert height.uncertainty_ratio == 0.015


@pytest.mark.parametrize("other", [(1.0, 1, Decimal("1.0"))])
def test_addition_only_with_measurements(other: Any) -> None:
    with pytest.raises(TypeError):
        Measurement(1 * Meter, 0.1) + other

    with pytest.raises(TypeError):
        other + Measurement(1 * Meter, 0.1)


def test_subtraction() -> None:
    # https://www.statisticshowto.com/statistics-basics/error-propagation/
    height = Measurement(2.00 * Meter, 0.03)
    waistband = Measurement(0.88 * Meter, 0.04)
    pant_length = height - waistband
    assert pant_length.measurand == 1.12 * Meter
    assert pant_length.uncertainty == 0.05 * Meter
    assert pant_length.uncertainty_ratio == approx(0.04464285)


def test_subtraction_from_quantity() -> None:
    # https://www.statisticshowto.com/statistics-basics/error-propagation/
    height = 2.00 * Meter
    waistband = Measurement(0.88 * Meter, 0.04)
    pant_length = height - waistband
    assert pant_length.measurand == 1.12 * Meter
    assert pant_length.uncertainty == 0.04 * Meter
    assert pant_length.uncertainty_ratio == approx(0.03571428)


def test_subtraction_of_quantity() -> None:
    # https://www.statisticshowto.com/statistics-basics/error-propagation/
    height = Measurement(2.00 * Meter, 0.03)
    waistband = 0.88 * Meter
    pant_length = height - waistband
    assert pant_length.measurand == 1.12 * Meter
    assert pant_length.uncertainty == 0.03 * Meter
    assert pant_length.uncertainty_ratio == approx(0.02678571)


@pytest.mark.parametrize("other", [(1.0, 1, Decimal("1.0"))])
def test_subtraction_only_with_measurements(other: Any) -> None:
    with pytest.raises(TypeError):
        Measurement(1 * Meter, 0.1) - other

    with pytest.raises(TypeError):
        other - Measurement(1 * Meter, 0.1)


def test_multiplication() -> None:
    # adapted from https://www.statisticshowto.com/statistics-basics/error-propagation/,
    # which absolutely has the wrong answer
    # Plugging the same problem into https://uncertaintycalculator.com/ agrees with
    # the results from measured
    length = Measurement(12 * Foot, 0.1)
    width = Measurement(10 * Foot, 0.1)
    height = Measurement(8 * Foot, 0.1)

    area = length * width
    assert area.measurand == 120 * Foot**2
    assert area.uncertainty == approximately(1.56204993518133 * Foot**2)
    assert area.uncertainty_ratio == approx(0.01301708)

    volume = area * height
    assert volume.measurand == 960 * Foot**3
    assert volume.uncertainty == approximately(17.3251262621662 * Foot**3)
    assert volume.uncertainty_ratio == approx(0.01804700)


def test_multiplication_by_quantity() -> None:
    length = Measurement(12 * Foot, 0.1)
    width = 10 * Foot

    area = length * width
    assert area.measurand == 120 * Foot**2
    area.uncertainty == 1.0 * Foot**2
    assert area.uncertainty_ratio == approx(0.00833333)


@pytest.mark.parametrize("other", [(1.0, 1, Decimal("1.0"))])
def test_multiplication_only_with_measurements(other: Any) -> None:
    with pytest.raises(TypeError):
        Measurement(1 * Meter, 0.1) * other

    with pytest.raises(TypeError):
        other * Measurement(1 * Meter, 0.1)


def test_division() -> None:
    volume = Measurement(960 * Foot**3, 17.3251262621662)
    length = Measurement(12 * Foot, 0.1)
    width = Measurement(10 * Foot, 0.1)

    area = volume / length
    assert area.measurand == 80 * Foot**2
    assert area.uncertainty == approximately(1.59024805891688 * Foot**2)
    assert area.uncertainty_ratio == approx(0.01987810)

    height = area / width
    assert height.measurand == 8 * Foot
    assert height.uncertainty == approximately(0.178013732304249 * Foot)
    assert height.uncertainty_ratio == approx(0.02225171)


def test_division_of_quantity() -> None:
    volume = 960 * Foot**3
    length = Measurement(12 * Foot, 0.1)

    area = volume / length
    assert area.measurand == 80 * Foot**2
    assert area.uncertainty == approximately(0.6666666666 * Foot**2)
    assert area.uncertainty_ratio == approx(0.00833333)


def test_division_by_quantity() -> None:
    volume = Measurement(960 * Foot**3, 17.3251262621662)
    length = 12 * Foot

    area = volume / length
    assert area.measurand == 80 * Foot**2
    assert area.uncertainty == approximately(1.443760522 * Foot**2)
    assert area.uncertainty_ratio == approx(0.01804700)


@pytest.mark.parametrize("other", [(1.0, 1, Decimal("1.0"))])
def test_division_only_with_measurements(other: Any) -> None:
    with pytest.raises(TypeError):
        Measurement(1 * Meter, 0.1) / other

    with pytest.raises(TypeError):
        other / Measurement(1 * Meter, 0.1)


def test_exponentation() -> None:
    side = Measurement(10 * Meter, 0.1)
    volume = side**3
    assert volume.measurand == 1000 * Meter**3
    assert volume.uncertainty == 30 * Meter**3
    assert volume.uncertainty_ratio == 0.03


@pytest.mark.parametrize("other", [(1 * Meter, 1.0, Decimal("1.0"))])
def test_exponentation_only_with_integers(other: Any) -> None:
    with pytest.raises(TypeError):
        Measurement(1 * Meter, 0.1) ** other

    with pytest.raises(TypeError):
        other ** Measurement(1 * Meter, 0.1)


@pytest.mark.parametrize(
    "left, right",
    [
        (Measurement(10 * Meter, 1), Measurement(10 * Meter, 1)),
        (Measurement(10 * Meter, 0.01), Measurement(10 * Meter, 0.01)),
        (Measurement(10 * Meter, 0.0000000001), Measurement(10 * Meter, 0.0000000001)),
        (Measurement(9.9 * Meter, 0.1), Measurement(10.1 * Meter, 0.1)),
        (Measurement(9.8 * Meter, 0.2), Measurement(10.2 * Meter, 0.2)),
        (Measurement(10 * Meter, 1), Measurement(11 * Meter, 1)),
        (Measurement(0 * Meter, 0.01), 0 * Meter),
        (Measurement(9.9 * Meter, 0.01), 9.9 * Meter),
        (Measurement(-9.9 * Meter, 0.01), -9.9 * Meter),
    ],
)
def test_equality(left: Measurement, right: Measurement) -> None:
    assert left == right
    assert right == left


@pytest.mark.parametrize(
    "left, right",
    [
        (Measurement(10 * Meter, 1), Measurement(13 * Meter, 1)),
        (Measurement(10 * Meter, 0.01), Measurement(10.2 * Meter, 0.01)),
        (Measurement(9.9 * Meter, 0.01), Measurement(10.1 * Meter, 0.01)),
        (Measurement(9.9 * Meter, 0.01), 10 * Meter),
        (Measurement(9.9 * Meter, 0.01), 9.9),
    ],
)
def test_unequal(left: Measurement, right: Measurement) -> None:
    assert left != right
    assert right != left


def test_simple_inequality() -> None:
    small = Measurement(10 * Meter, 1)
    medium = Measurement(20 * Meter, 1)
    large = Measurement(30 * Meter, 1)

    assert small <= medium < large
    assert large >= medium > small


def test_inequalities_based_on_uncertainty() -> None:
    assert Measurement(10 * Meter, 1.0) < Measurement(10 * Meter, 0.1)
    assert Measurement(10 * Meter, 1.0) < Measurement(9.9 * Meter, 0.1)
    assert Measurement(10 * Meter, 1.0) > Measurement(10 * Meter, 0.5)
    assert Measurement(10 * Meter, 1.0) > Measurement(10.1 * Meter, 0.2)


def test_inequalities_coerce_to_measurements() -> None:
    tiny = -5 * Decibel[1 * Meter]
    lesser = 8 * Meter
    length = Measurement(10 * Meter, 1)
    greater = 20 * Meter
    massive = 100 * Decibel[1 * Meter]

    assert tiny < length < massive
    assert tiny <= length <= massive
    assert massive > length > tiny
    assert massive >= length >= tiny

    assert tiny < lesser < length < greater < massive
    assert tiny <= lesser <= length <= greater <= massive
    assert massive > greater > length > lesser > tiny
    assert massive >= greater >= length >= lesser >= tiny


def test_inequalities_only_with_measurements_or_quantities() -> None:
    length = Measurement(10 * Meter, 1)

    with pytest.raises(TypeError):
        length < 100
    with pytest.raises(TypeError):
        length <= 100
    with pytest.raises(TypeError):
        length > 1
    with pytest.raises(TypeError):
        length >= 1


def test_repr() -> None:
    assert repr(Measurement(100 * Meter, 0.01)) == (
        "Measurement("
        "measurand=Quantity(magnitude=100, unit=Unit.named('meter')), "
        "uncertainty=0.01"
        ")"
    )


def test_str() -> None:
    assert str(Measurement(100 * Meter, 0.01)) == "100±0.01 m"


@pytest.mark.parametrize(
    "template, expected",
    [
        ("{speed}", "1234.5678±0.234 m⋅s⁻¹"),
        ("{speed:+}", "1234.5678±0.234 m⋅s⁻¹"),
        ("{speed:+.2f}", "1234.5678±0.23 m⋅s⁻¹"),
        ("{speed:+:.2f}", "1234.57±0.234 m⋅s⁻¹"),
        ("{speed:+:.2f:/}", "1234.57±0.234 m/s"),
        ("{speed:+.2f:.2f:/}", "1234.57±0.23 m/s"),
        ("{speed:%}", "1234.5678±0.02% m⋅s⁻¹"),  # default to 2 decimal places
        ("{speed:%.3f}", "1234.5678±0.019% m⋅s⁻¹"),
        ("{speed:%.3f::/}", "1234.5678±0.019% m/s"),
    ],
)
def test_formatting(template: str, expected: str) -> None:
    speed = Measurement(1234.5678 * Meter / Second, 0.234)
    assert template.format(speed=speed) == expected


def test_unrecognized_format_specifier() -> None:
    speed = Measurement(1234.5678 * Meter / Second, 0.234)
    with pytest.raises(ValueError, match="Unrecognized uncertainty style '🤷'"):
        speed.__format__("🤷")


@pytest.mark.parametrize(
    "quantity",
    [
        -10 * Meter,
        -10.0 * Meter,
        0 * Meter,
        0.0 * Meter,
        10 * Meter,
        10.0 * Meter,
    ],
)
def test_quantities_always_approximate_themselves(quantity: Quantity) -> None:
    assert quantity == approximately(quantity)



"""
Defines both the [fundamental _dimensionless_ constants][1], as well as the
[foundational _dimensioned_ physical constants][2] of the universe.

[1]: https://en.wikipedia.org/wiki/Dimensionless_physical_constant
[2]: https://en.wikipedia.org/wiki/Physical_constant

Attributes: Physical constants

    c (Quantity): The [speed that light travels unimpeded in a vacuum][1]

        [1]: https://en.wikipedia.org/wiki/Speed_of_light

    G (Quantity): The "strength" of the [gravitational force][1]

        [1]: https://en.wikipedia.org/wiki/Gravitational_constant

    h (Quantity): The [Planck constant][1]

        [1]: https://en.wikipedia.org/wiki/Planck_constant

    ℏ (Quantity): The ["reduced" Planck constant][1]

        [1]: https://en.wikipedia.org/wiki/Planck_constant#Reduced_Planck_constant

    k (Quantity): The [Boltzmann constant][1]

        [1]: https://en.wikipedia.org/wiki/Boltzmann_constant

    e (Quantity): The [elementary charge][1]

        [1]: https://en.wikipedia.org/wiki/Elementary_charge

    ε0 (Quantity): [Vacuum permittivity][1], the capability of an electric field to
        permeate a vacuum

        [1]: https://en.wikipedia.org/wiki/Vacuum_permittivity

    mₑ (Quantity): The [mass of an electron][1]

        [1]: https://en.wikipedia.org/wiki/Electron_mass

    kₑ (Quantity): Coulomb's [electric force constants][1]

        [1]: https://en.wikipedia.org/wiki/Coulomb_constant

    α (Quantity): The [fine-structure constant][1], quantifying the strength of
        electromagnetic interactions

        [1]: https://en.wikipedia.org/wiki/Fine-structure_constant


"""

from measured.geometry import π
from measured.si import Coulomb, Farad, Hertz, Joule, Kelvin, Kilogram, Meter, Second

# https://en.wikipedia.org/wiki/Speed_of_light
c = 299792458 * Meter / Second

# https://en.wikipedia.org/wiki/Gravitational_constant
G = 6.67430e-11 * Meter**3 / (Kilogram * Second**2)

# https://en.wikipedia.org/wiki/Planck_constant#Value
h = 6.62607015e-34 * Joule / Hertz

# The "reduced" Planck constant, "h-bar"
ℏ = h / (2 * π)

# https://en.wikipedia.org/wiki/Boltzmann_constant
k = 1.380649e-23 * Joule / Kelvin

# https://en.wikipedia.org/wiki/Elementary_charge
e = 1.602176634e-19 * Coulomb

# https://en.wikipedia.org/wiki/Vacuum_permittivity
ε0 = 8.8541878128e-12 * Farad / Meter

# https://en.wikipedia.org/wiki/Electron_mass
mₑ = 9.1093837015e-31 * Kilogram

# https://en.wikipedia.org/wiki/Coulomb_constant
kₑ = (4 * π * ε0) ** -1

# https://en.wikipedia.org/wiki/Fine-structure_constant
α = (e**2) / (4 * π * ε0 * ħ * c)

# https://en.wikipedia.org/wiki/Standard_gravity
gₙ = 9.80665 * Meter / Second**2



"""
Defines several systems of [natural units][1] of measurement, derived entirely from
[universal physical constants][2], as defined in [`measured.physics`](../physics).
These units simplify the expression of concepts in cosmology and particle physics.

[1]: https://en.wikipedia.org/wiki/Natural_units

[2]: https://en.wikipedia.org/wiki/Physical_constant


## Planck Units

The [Planck units][1] are defined such that `c`, `G`, `ℏ`, and `k` equal `1`.

[1]: https://en.wikipedia.org/wiki/Planck_units


Attributes: Planck units

    PlanckLength (Unit):

    PlanckMass (Unit):

    PlanckTime (Unit):

    PlanckTemperature (Unit):

## Stoney units

The [Stoney units][1] are defined such that `c`, `G`, `kₑ` and `e` equal `1`.

[1]: https://en.wikipedia.org/wiki/Stoney_units

Attributes: Stoney units

    StoneyLength (Unit):

    StoneyMass (Unit):

    StoneyTime (Unit):

    StoneyCharge (Unit):

## Hartree atomic units

The [Atomic units][1] are defined such that `e`, `mₑ`, `ℏ`, and `kₑ` equal `1`.

[1]: https://en.wikipedia.org/wiki/Hartree_atomic_units

Attributes: Hartree atomic units

    AtomicLength (Unit):

    AtomicMass (Unit):

    AtomicTime (Unit):

    AtomicCharge (Unit):

## Naturral units

The [Natural units][1] are defined such that `c`, `mₑ`, `ℏ`, and `ε0` equal `1`.

[1]: https://en.wikipedia.org/wiki/Natural_units

Attributes: Natural units

    NaturalLength (Unit):

    NaturalMass (Unit):

    NaturalTime (Unit):

    NaturalCharge (Unit):

"""

from measured import Charge, Length, Mass, Temperature, Time
from measured.geometry import π
from measured.physics import G, c, e, k, kₑ, mₑ, ε0, ℏ

PlanckLength = Length.unit("planck length", "lₚ")
PlanckLength.equals(((ℏ * G) / c**3).root(2))

PlanckMass = Mass.unit("planck mass", "mₚ")
PlanckMass.equals(((ℏ * c) / G).root(2))

PlanckTime = Time.unit("planck time", "tₚ")
PlanckTime.equals(((ℏ * G) / c**5).root(2))

PlanckTemperature = Temperature.unit("planck temperature", "Tₚ")
PlanckTemperature.equals(((ℏ * c**5) / (G * k**2)).root(2))


# https://en.wikipedia.org/wiki/Stoney_units

StoneyLength = Length.unit("stoney length", "lₛ")
StoneyLength.equals(((G * kₑ * e**2) / c**4).root(2))

StoneyMass = Mass.unit("stoney mass", "mₛ")
StoneyMass.equals(((kₑ * e**2) / G).root(2))

StoneyTime = Time.unit("stoney time", "tₛ")
StoneyTime.equals(((G * kₑ * e**2) / c**6).root(2))

StoneyCharge = Charge.unit("stoney charge", "qₛ")
StoneyCharge.equals(e)


# https://en.wikipedia.org/wiki/Natural_units#Atomic_units

AtomicLength = Length.unit("atomic length", "lₐ")
AtomicLength.equals(((4 * π * ε0) * ℏ**2) / (mₑ * e**2))

AtomicMass = Mass.unit("atomic mass", "mₐ")
AtomicMass.equals(mₑ)

AtomicTime = Time.unit("atomic time", "tₐ")
AtomicTime.equals(((4 * π * ε0) ** 2 * ℏ**3) / (mₑ * e**4))

AtomicCharge = Charge.unit("atomic charge", "qₐ")
AtomicCharge.equals(e)


# https://en.wikipedia.org/wiki/Natural_units#Natural_units_(particle_and_atomic_physics)

NaturalLength = Length.unit("natural length", "lₙ")
NaturalLength.equals((ℏ / (mₑ * c)))

NaturalMass = Mass.unit("natural mass", "mₙ")
NaturalMass.equals(mₑ)

NaturalTime = Time.unit("natural time", "tₙ")
NaturalTime.equals((ℏ / (mₑ * c**2)))

NaturalCharge = Charge.unit("natural charge", "qₙ")
NaturalCharge.equals((ε0 * ℏ * c).root(2))



from measured import Length
from measured.iec import RackUnit
from measured.us import Inch


def test_rack_unit() -> None:
    assert RackUnit.dimension is Length
    assert RackUnit.name == "rack unit"
    assert RackUnit.symbol == "U"
    assert 1 * RackUnit == 1.75 * Inch



import pytest

from measured import Quantity, approximately
from measured.si import Celsius, Kelvin, Kilo, Milli


@pytest.mark.parametrize(
    "kelvin, celsius",
    [
        (0 * Kelvin, -273.15 * Celsius),
        (273.15 * Kelvin, 0 * Celsius),
        (373.15 * Kelvin, 100 * Celsius),
        (10273.15 * Kelvin, 10 * Kilo * Celsius),
        (1 * Kilo * Kelvin, 726.85 * Celsius),
    ],
)
def test_kelvin_equals_celsius(kelvin: Quantity, celsius: Quantity) -> None:
    assert kelvin == approximately(celsius, 0)
    assert celsius == approximately(kelvin, 0)


@pytest.mark.parametrize(
    "kelvin, celsius",
    [
        (273.16 * Kelvin, 10 * Milli * Celsius),
        (283150.0 * Milli * Kelvin, 10 * Celsius),
    ],
)
def test_kelvin_approximates_celsius(kelvin: Quantity, celsius: Quantity) -> None:
    assert kelvin == approximately(celsius)
    assert celsius == approximately(kelvin)



from io import StringIO
from reprlib import Repr

import pytest

pytest.importorskip("IPython")

# flake8: noqa: E402 (imports not at top of file)

from typing import Union
from xml.etree import ElementTree

from IPython.lib.pretty import RepresentationPrinter

from measured import (
    Area,
    Decibel,
    Dimension,
    Frequency,
    Length,
    Logarithm,
    Measurement,
    Neper,
    Number,
    Prefix,
    Quantity,
    Time,
    Unit,
    Volume,
    VolumetricFlow,
)
from measured.formatting import superscript
from measured.si import Hertz, Kilo, Mega, Meter, Milli, Ohm, Second, Watt


@pytest.mark.parametrize(
    "exponent, superscripted",
    [
        (-1234567890, "⁻¹²³⁴⁵⁶⁷⁸⁹⁰"),
        (-123.456, "⁻¹²³.⁴⁵⁶"),
        (-123, "⁻¹²³"),
        (-23, "⁻²³"),
        (-3, "⁻³"),
        (-1.1, "⁻¹.¹"),
        (-1.0, "⁻¹.⁰"),
        (-1, "⁻¹"),
        (-0.9, "⁻⁰.⁹"),
        (0, "⁰"),
        (1, ""),
        (0.9, "⁰.⁹"),
        (1.0, ""),
        (1.1, "¹.¹"),
        (2, "²"),
        (23, "²³"),
        (345, "³⁴⁵"),
        (123.456, "¹²³.⁴⁵⁶"),
        (1234567890, "¹²³⁴⁵⁶⁷⁸⁹⁰"),
    ],
)
def test_superscripts(exponent: int, superscripted: str) -> None:
    assert superscript(exponent) == superscripted


def test_formatting_prefixes_simplifies() -> None:
    # This is a little surprising, but what's happening here is that
    # 5 meter per kilosecond is getting reduced to 5 millimeter per second
    assert str((5 * Meter) / (Kilo * Second)) == "5 mm⋅s⁻¹"


def test_do_the_best_we_can_with_odd_prefixes() -> None:
    kilo_plus_one = Prefix(10, 4)
    assert str(5 * (kilo_plus_one * Meter)) == "5 10⁴m"


@pytest.mark.parametrize(
    "dimension, string",
    [
        (Number, "1"),
        (Length, "L"),
        (Area, "L²"),
        (Volume, "L³"),
        (Frequency, "T⁻¹"),
    ],
)
def test_strings_of_dimensions(dimension: Dimension, string: str) -> None:
    assert str(dimension) == string


@pytest.mark.parametrize(
    "unit, string",
    [
        (Meter, "m"),
        (Meter**2, "m²"),
        (Meter**3, "m³"),
        (Meter**3 / Second**2, "m³⋅s⁻²"),
        (Hertz, "Hz"),
        (Hertz**2, "s⁻²"),
        (Decibel[1 * Meter], "dB of 1 m"),
        (Logarithm(3), "1 log3(x/x₀)"),
    ],
)
def test_strings_of_units(unit: Unit, string: str) -> None:
    assert str(unit) == string


@pytest.mark.parametrize(
    "quantity, string",
    [
        (5 * Meter, "5 m"),
        (5.1 * Meter**2, "5.1 m²"),
        (5 * (Kilo * Meter), "5 km"),
        (5.1 * (Kilo * (Meter**2) / Second), "5.1 km²⋅s⁻¹"),
        (5.1 * (Kilo * Meter**2) / Second, "5.1 km²⋅s⁻¹"),
        (5.1 * (Kilo * Meter) ** 2 / Second, "5.1 Mm²⋅s⁻¹"),
        (5.1 * (Mega * Meter**-1), "5.1 μm⁻¹"),
    ],
)
def test_strings_of_quantities(quantity: Quantity, string: str) -> None:
    assert str(quantity) == string


@pytest.mark.parametrize(
    "quantity, template, string",
    [
        (5 * Meter, "{quantity}", "5 m"),
        (5.123 * Meter, "{quantity:.2f:/}", "5.12 m"),
        (5.123 * Meter / Second, "{quantity:.2f}", "5.12 m⋅s⁻¹"),
        (5.123 * Meter / Second, "{quantity:.2f:}", "5.12 m⋅s⁻¹"),
        (5.123 * Meter / Second, "{quantity:.2f:/}", "5.12 m/s"),
        (5 * Ohm, "{quantity}", "5 Ω"),
        (5 * Ohm, "{quantity::/}", "5 m²⋅kg/s⋅C²"),
        (5 * Milli * Ohm, "{quantity::/}", "0.005 m²⋅kg/s⋅C²"),
    ],
)
def test_format_specifiers(quantity: Quantity, template: str, string: str) -> None:
    assert template.format(quantity=quantity) == string


def test_unrecognized_unit_format_specifier() -> None:
    with pytest.raises(ValueError):
        "{quantity::nope}".format(quantity=5 * Meter)


Formattable = Union[Dimension, Prefix, Unit, Quantity, Measurement]


@pytest.fixture
def pretty() -> RepresentationPrinter:
    return RepresentationPrinter(output=StringIO())


@pytest.mark.parametrize(
    "formattable",
    [
        Length,
        Area,
        VolumetricFlow,
        Kilo,
        Prefix(10, 4),
        Meter,
        Kilo * Meter,
        Prefix(10, 4) * Meter,
        Ohm,
        5 * Meter,
        Measurement(5 * Meter, 0.1),
        Decibel,
        Neper,
        Decibel[1 * Watt],
        30 * Decibel[1 * Watt],
    ],
)
def test_pretty_repr_includes_string_and_repr_of_self(
    formattable: Formattable, pretty: RepresentationPrinter
) -> None:
    formattable._repr_pretty_(pretty, False)
    assert str(formattable) in pretty.output.getvalue()
    assert repr(formattable) in pretty.output.getvalue()


@pytest.mark.parametrize(
    "formattable, specifier",
    [
        (Kilo * Meter / Second, "/"),
        (5 * Meter / Second, ":/"),
        (Measurement(5 * Meter / Second, 0.1), "::/"),
    ],
)
def test_pretty_repr_includes_string_of_self(
    formattable: Formattable, specifier: str, pretty: RepresentationPrinter
) -> None:
    formattable._repr_pretty_(pretty, False)
    assert formattable.__format__(specifier) in pretty.output.getvalue()


@pytest.mark.parametrize(
    "formattable",
    [
        Length,
        Area,
        VolumetricFlow,
        Kilo,
        Prefix(10, 4),
        Meter,
        Kilo * Meter,
        Prefix(10, 4) * Meter,
        Ohm,
        Kilo * Meter / Second,
        5 * Meter,
        5 * Meter / Second,
        Measurement(5 * Meter, 0.1),
        Measurement(5 * Meter / Second, 0.1),
        Neper,
        Decibel[1 * Watt],
        30 * Decibel[1 * Watt],
    ],
)
def test_html_is_mathml(formattable: Formattable) -> None:
    math = ElementTree.fromstring(formattable._repr_html_())
    assert math.tag == "math"


@pytest.mark.parametrize(
    "formattable",
    [
        Length / Time,
        Meter / Second,
    ],
)
def test_mathml_root_is_fraction(formattable: Formattable) -> None:
    math = ElementTree.fromstring(formattable._repr_html_())
    assert math[0].tag == "mfrac"


@pytest.mark.parametrize(
    "formattable",
    [
        Kilo,
        Meter,
        Hertz,
        Neper,
        Decibel[1 * Watt],
    ],
)
def test_mathml_root_is_identifier(formattable: Formattable) -> None:
    math = ElementTree.fromstring(formattable._repr_html_())
    assert math[0].tag == "mi"


@pytest.mark.parametrize(
    "formattable",
    [
        Length,
        Length * Time,
        (Kilo * Meter),
        Meter * Second,
        Logarithm(3),
        Decibel[1 * Meter],  # a dB that we wouldn't have a symbol for
        Neper[1 * Meter],
        30 * Decibel[1 * Watt],
    ],
)
def test_mathml_root_is_subexpression(formattable: Formattable) -> None:
    math = ElementTree.fromstring(formattable._repr_html_())
    assert math[0].tag == "mrow"



from measured import Length
from measured.si import Hertz, Meter

# ISO 2848:1984 - Building construction — Modular coordination — Principles and rules
# https://www.iso.org/standard/7846.html

BasicModule = Length.unit("basic module", "basic-module")
BasicModule.equals(0.1 * Meter)

MetricInch = Length.unit("metric inch", "metric-foot")
MetricInch.equals(0.25 * BasicModule)

MetricFoot = Length.unit("metric foot", "metric-inch")
MetricInch.equals(3 * BasicModule)


# ISO 16:1975 Acoustics, Standard tuning frequency (Standard musical pitch)
# http://www.iso.org/iso/iso_catalogue/catalogue_tc/catalogue_detail.htm?csnumber=3601

A440 = 440 * Hertz



import pytest
from pytest import approx

from measured import Decibel, Level, Pressure, Quantity, approximately
from measured.acoustics import ABSOLUTE_THRESHOLD_OF_HEARING, dBSPL, dBSWL
from measured.geometry import π
from measured.si import Meter, Micro, Pascal


def test_auditory_threshold() -> None:
    assert ABSOLUTE_THRESHOLD_OF_HEARING.unit.dimension == Pressure
    assert ABSOLUTE_THRESHOLD_OF_HEARING == 20 * Micro * Pascal
    assert ABSOLUTE_THRESHOLD_OF_HEARING == 0 * dBSPL


# https://en.wikipedia.org/wiki/Sound_pressure#Examples_of_sound_pressure
@pytest.mark.parametrize(
    "pressure, level, within",
    [
        # Microsoft's A-weighted anechoic chamber
        (1.9e-6 * Pascal, -20.35 * dBSPL, 0.02),
        # threshold of human hearing
        (ABSOLUTE_THRESHOLD_OF_HEARING, 0 * dBSPL, 1e-7),
        # Jackhammer at 1 meter
        (2 * Pascal, 100 * dBSPL, 1e-7),
        # Chainsaw at 1 meter
        (6.32 * Pascal, 110 * dBSPL, 1e-3),
        # Jet Engine at 1 meter
        (632 * Pascal, 150 * dBSPL, 1e-3),
    ],
)
def test_examples_of_sound_pressure(
    pressure: Quantity, level: Level, within: float
) -> None:
    assert pressure == approximately(level, within=within)
    assert level == approximately(pressure, within=within)


def test_sound_power_from_sound_pressure() -> None:
    # https://en.wikipedia.org/wiki/Sound_power#Relationship_with_sound_pressure_level
    # https://en.wikipedia.org/wiki/Sound_pressure#Sound_pressure_level

    # cross-checked with
    # https://www.wkcgroup.com/tools-room/sound-power-sound-pressure-converter/

    distance = 2 * Meter
    area = 4 * π * distance**2
    assert area == approximately(50.265482 * Meter**2)

    surface_area_level = area.level(Decibel[1 * Meter**2])
    assert surface_area_level == approximately(17.01269855 * Decibel[1 * Meter**2])

    pressure = 50 * dBSPL
    power = 67.0127 * dBSWL

    assert pressure.magnitude + surface_area_level.magnitude == approx(power.magnitude)



from math import inf, nan

import pytest
from hypothesis import assume, example, given
from hypothesis.strategies import floats, integers

from measured import Numeric, Quantity, Unit, approximately, systems  # noqa: F401
from measured.hypothesis import units
from measured.iec import Byte, Kibi
from measured.parsing import ParseError
from measured.si import (
    Ampere,
    Gram,
    Hertz,
    Kilo,
    Kilogram,
    Liter,
    Mega,
    Meter,
    Micro,
    Ohm,
    Second,
)


def test_unit_symbols_should_not_have_spaces() -> None:
    with pytest.raises(ValueError, match="spaces"):
        Meter.alias(symbol="the metre")


@given(unit=units())
def test_each_unit_roundtrips(unit: Unit) -> None:
    assume(unit is not (Kilo * Gram))  # see test below
    assert Unit.parse(str(unit)) is unit


def test_kilogram_converts_to_kilogram() -> None:
    # This test is a special case for Kilo*Gram versus Kilogram, which measured will
    # always parse to Kilogram
    assert Unit.parse(str(Kilogram)) is Kilogram
    assert Unit.parse(str(Kilo * Gram)) is Kilogram


@given(
    magnitude=integers(min_value=-1000000000000, max_value=1000000000000),
    unit=units(),
)
@example(magnitude=1, unit=Liter)
@example(magnitude=1, unit=Mega * (Meter**-1))
def test_small_integer_quantities_parse_exactly(magnitude: Numeric, unit: Unit) -> None:
    assert Quantity.parse(str(magnitude * unit)) == magnitude * unit


@given(magnitude=integers(), unit=units())
@example(magnitude=1, unit=Liter)
def test_large_integer_quantities_are_parsable(magnitude: Numeric, unit: Unit) -> None:
    assert Quantity.parse(str(magnitude * unit)) == approximately(magnitude * unit)


@given(magnitude=floats(allow_infinity=False, allow_nan=False), unit=units())
@example(magnitude=1.0, unit=Liter)
def test_float_quantities_are_parsable(magnitude: Numeric, unit: Unit) -> None:
    assert Quantity.parse(str(magnitude * unit)) == magnitude * unit


@given(unit=units())
@pytest.mark.parametrize("magnitude", [nan, inf, -inf])
def test_unparsable_floats(magnitude: float, unit: Unit) -> None:
    with pytest.raises(ParseError):
        Quantity.parse(str(magnitude * unit))


@pytest.mark.parametrize(
    "string",
    [
        "fleebles",
        "queeble²",
        ",,,",
        "",
        " ",
    ],
)
def test_unit_that_should_not_parse(string: str) -> None:
    with pytest.raises((KeyError, ParseError)):
        Unit.parse(string)


@pytest.mark.parametrize(
    "string, unit",
    [
        ("m", Meter),
        ("meter", Meter),
        ("Hz", Hertz),
        ("hertz", Hertz),
        ("Ω", Ohm),
        ("ohm", Ohm),
    ],
)
def test_single_unit(string: str, unit: Unit) -> None:
    assert Unit.parse(string) is unit


@pytest.mark.parametrize(
    "string, unit",
    [
        ("m²", Meter**2),
        ("meter²", Meter**2),
        ("m²²²", Meter**222),
        ("m⁻¹", Meter**-1),
        ("m^2", Meter**2),
        ("m^222", Meter**222),
        ("m^-1", Meter**-1),
    ],
)
def test_unit_with_exponent(string: str, unit: Unit) -> None:
    assert Unit.parse(string) is unit


@pytest.mark.parametrize(
    "string, unit",
    [
        ("m²A³", Meter**2 * Ampere**3),
        ("m²⋅A³", Meter**2 * Ampere**3),
        ("m²*A³", Meter**2 * Ampere**3),
        ("m²A³s⁻³", Meter**2 * Ampere**3 * Second**-3),
        ("m²⋅A³⋅s^-3", Meter**2 * Ampere**3 * Second**-3),
        ("m²*A³*s^-3", Meter**2 * Ampere**3 * Second**-3),
    ],
)
def test_unit_series(string: str, unit: Unit) -> None:
    assert Unit.parse(string) is unit


@pytest.mark.parametrize(
    "string, unit",
    [
        ("m/s", Meter / Second),
        ("m²A³/s^3", (Meter**2 * Ampere**3) / Second**3),
        ("m²⋅A³/s^3", (Meter**2 * Ampere**3) / Second**3),
        ("m²*A³/s^3", (Meter**2 * Ampere**3) / Second**3),
    ],
)
def test_unit_ratio(string: str, unit: Unit) -> None:
    assert Unit.parse(string) == unit


@pytest.mark.parametrize(
    "string",
    [
        "m",
        "5 zeebles",
    ],
)
def test_quantities_that_should_not_parse(string: str) -> None:
    with pytest.raises((KeyError, ParseError)):
        Quantity.parse(string)


@pytest.mark.parametrize(
    "string, quantity",
    [
        ("5m", 5 * Meter),
        ("5 m", 5 * Meter),
        ("5.1 m", 5.1 * Meter),
        ("5 Hz", 5 * Hertz),
        ("5.1 Hz", 5.1 * Hertz),
        ("5 Ω", 5 * Ohm),
        ("5.1 Ω", 5.1 * Ohm),
    ],
)
def test_quantity_with_single_unit(string: str, quantity: Quantity) -> None:
    assert Quantity.parse(string) == quantity


@pytest.mark.parametrize(
    "string, quantity",
    [
        ("5 m²", 5 * Meter**2),
        ("5.1 m²", 5.1 * Meter**2),
        ("5 m²²²", 5 * Meter**222),
        ("5 m⁻¹", 5 * Meter**-1),
        ("5 m^2", 5 * Meter**2),
        ("5.1 m^2", 5.1 * Meter**2),
        ("5 m^222", 5 * Meter**222),
        ("5 m^-1", 5 * Meter**-1),
    ],
)
def test_quantity_with_unit_exponent(string: str, quantity: Quantity) -> None:
    assert Quantity.parse(string) == quantity


@pytest.mark.parametrize(
    "string, quantity",
    [
        ("5 m²A³", 5 * Meter**2 * Ampere**3),
        ("5 m²⋅A³", 5 * Meter**2 * Ampere**3),
        ("5 m²*A³", 5 * Meter**2 * Ampere**3),
        ("5 m²A³s⁻³", 5 * Meter**2 * Ampere**3 * Second**-3),
        ("5 m²⋅A³⋅s^-3", 5 * Meter**2 * Ampere**3 * Second**-3),
        ("5 m²*A³*s^-3", 5 * Meter**2 * Ampere**3 * Second**-3),
    ],
)
def test_quantity_with_unit_series(string: str, quantity: Quantity) -> None:
    assert Quantity.parse(string) == quantity


@pytest.mark.parametrize(
    "string, quantity",
    [
        ("5 m/s", 5 * Meter / Second),
        ("5 m²A³/s^3", 5 * (Meter**2 * Ampere**3) / Second**3),
        ("5 m²⋅A³/s^3", 5 * (Meter**2 * Ampere**3) / Second**3),
        ("5 m²*A³/s^3", 5 * (Meter**2 * Ampere**3) / Second**3),
    ],
)
def test_quantity_with_unit_ratio(string: str, quantity: Quantity) -> None:
    assert Quantity.parse(string) == quantity


@pytest.mark.parametrize(
    "string, quantity",
    [
        ("5 km", 5 * Kilo * Meter),
        ("5 KiB", 5 * Kibi * Byte),
        ("5 μm", 5 * Micro * Meter),
    ],
)
def test_quantity_with_unit_prefixes(string: str, quantity: Quantity) -> None:
    assert Quantity.parse(string) == quantity



"""
Defines the [Troy weights][1] and their conversions to SI

[1]: https://en.wikipedia.org/wiki/Troy_weight

Attributes: Units of mass ("weight")

    Grain (Unit): The same `Grain` as the Avoirdupois system

    Pennyweight (Unit): 24 grains

    Ounce (Unit): 20 troy pennyweights

    Pound (Unit): 12 troy ounces

"""

from measured import Mass, avoirdupois
from measured.si import Gram

# https://en.wikipedia.org/wiki/Troy_weight#Troy_grain
# There is no specific 'troy grain'. All Imperial systems use the same measure of mass
# called a grain
Grain = avoirdupois.Grain

Pennyweight = Mass.unit("pennyweight", "dwt")
Pennyweight.equals(24 * Grain)

Ounce = Mass.unit("troy ounce", "oz-t")
Ounce.equals(480 * Grain)
Ounce.equals(20 * Pennyweight)
Ounce.equals(31.10348 * Gram)

Pound = Mass.unit("troy pound", "lb-t")
Pound.equals(12 * Ounce)
Pound.equals(240 * Pennyweight)
Pound.equals(373.24172 * Gram)
